const priorities = {
  CRITICAL: 0,
  HIGH: 1,
  LOW: 2
};
const validPriorities = new Set([0, 1, 2]);

const queues = {
  [priorities.CRITICAL]: [],
  [priorities.HIGH]: [],
  [priorities.LOW]: []
};

function validatePriority(priority) {
  if (!validPriorities.has(priority)) {
    throw new Error(`Invalid queue priority: ${priority}.`);
  }
  return priority;
}

let tickTask; // tick means the next microtask
let rafTask; // raf means the next animation frame (requestAnimationFrame)
let ricTask; // ric means the next idle perdiod (requestIdleCallback)

const currentTick = Promise.resolve();

function getRaf() {
  return typeof requestAnimationFrame === 'function' ? requestAnimationFrame : setTimeout;
}

function getRic() {
  return typeof requestIdleCallback === 'function' ? requestIdleCallback : getRaf();
}

// schedule a tick task, if it is not yet scheduled
function nextTick(task) {
  if (!tickTask) {
    tickTask = task;
    currentTick.then(runTickTask);
  }
}

function runTickTask() {
  const task = tickTask;
  // set the task to undefined BEFORE calling it
  // this allows it to re-schedule itself for a later time
  tickTask = undefined;
  task();
}

// schedule a raf task, if it is not yet scheduled
function nextAnimationFrame(task) {
  if (!rafTask) {
    rafTask = task;
    const raf = getRaf();
    raf(runRafTask);
  }
}

function runRafTask() {
  const task = rafTask;
  // set the task to undefined BEFORE calling it
  // this allows it to re-schedule itself for a later time
  rafTask = undefined;
  task();
}

// schedule a ric task, if it is not yet scheduled
function nextIdlePeriod(task) {
  if (!ricTask) {
    ricTask = task;
    const ric = getRic();
    ric(runRicTask);
  }
}

function runRicTask() {
  // do not run ric task if there are pending raf tasks
  // let the raf tasks execute first and schedule the ric task for later
  if (!rafTask) {
    const task = ricTask;
    // set the task to undefined BEFORE calling it
    // this allows it to re-schedule itself for a later time
    ricTask = undefined;
    task();
  } else {
    const ric = getRic();
    ric(runRicTask);
  }
}

const TARGET_FPS = 60;
const TARGET_INTERVAL = 1000 / TARGET_FPS;

function queueTaskProcessing(priority) {
  if (priority === priorities.CRITICAL) {
    nextTick(runQueuedCriticalTasks);
  } else if (priority === priorities.HIGH) {
    nextAnimationFrame(runQueuedHighTasks);
  } else {
    nextIdlePeriod(runQueuedLowTasks);
  }
}

function runQueuedCriticalTasks() {
  // critical tasks must all execute before the next frame
  const criticalQueues = queues[priorities.CRITICAL];
  criticalQueues.forEach(processCriticalQueue);
}

function processCriticalQueue(queue) {
  queue.forEach(runTask);
  queue.clear();
}

function runQueuedHighTasks() {
  const startTime = performance.now();
  const isEmpty = processIdleQueues(priorities.HIGH, startTime);
  // there are more tasks to run in the next cycle
  if (!isEmpty) {
    nextAnimationFrame(runQueuedHighTasks);
  }
}

function runQueuedLowTasks() {
  const startTime = performance.now();
  const isEmpty = processIdleQueues(priorities.LOW, startTime);
  // there are more tasks to run in the next cycle
  if (!isEmpty) {
    nextIdlePeriod(runQueuedLowTasks);
  }
}

function processIdleQueues(priority, startTime) {
  const idleQueues = queues[priority];
  let isEmpty = true;

  // if a queue is not empty after processing, it means we have no more time
  // the loop whould stop in this case
  for (let i = 0; isEmpty && i < idleQueues.length; i++) {
    const queue = idleQueues.shift();
    isEmpty = isEmpty && processIdleQueue(queue, startTime);
    idleQueues.push(queue);
  }
  return isEmpty;
}

function processIdleQueue(queue, startTime) {
  const iterator = queue[Symbol.iterator]();
  let task = iterator.next();
  while (performance.now() - startTime < TARGET_INTERVAL) {
    if (task.done) {
      return true;
    }
    runTask(task.value);
    queue.delete(task.value);
    task = iterator.next();
  }
}

function runTask(task) {
  task();
}

const QUEUE = Symbol('task queue');

class Queue {
  constructor(priority) {
    this.priority = validatePriority(priority);
    this[QUEUE] = new Set();
    queues[this.priority].push(this[QUEUE]);
  }

  has(task) {
    return this[QUEUE].has(task);
  }

  add(task) {
    if (typeof task !== 'function') {
      throw new TypeError(`${task} can not be added to the queue. Only functions can be added.`);
    }
    const queue = this[QUEUE];
    queue.add(task);
    queueTaskProcessing(this.priority);
  }

  delete(task) {
    this[QUEUE].delete(task);
  }

  start() {
    const queue = this[QUEUE];
    const priorityQueues = queues[this.priority];
    if (priorityQueues.indexOf(queue) === -1) {
      priorityQueues.push(queue);
    }
    queueTaskProcessing(this.priority);
  }

  stop() {
    const queue = this[QUEUE];
    const priorityQueues = queues[this.priority];
    const index = priorityQueues.indexOf(queue);
    if (index !== -1) {
      priorityQueues.splice(index, 1);
    }
  }

  clear() {
    this[QUEUE].clear();
  }

  get size() {
    return this[QUEUE].size;
  }

  process() {
    const queue = this[QUEUE];
    queue.forEach(runTask);
    queue.clear();
  }

  processing() {
    const queue = this[QUEUE];
    return new Promise(resolve => {
      if (queue.size === 0) {
        resolve();
      } else {
        queue.add(resolve);
      }
    });
  }
}

export { Queue, priorities };
