(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(factory((global.queue = {})));
}(this, (function (exports) { 'use strict';

var priorities = {
  CRITICAL: 0,
  HIGH: 1,
  LOW: 2
};
var validPriorities = new Set([0, 1, 2]);

var queues = {};
queues[priorities.CRITICAL] = [];
queues[priorities.HIGH] = [];
queues[priorities.LOW] = [];

function validatePriority(priority) {
  if (!validPriorities.has(priority)) {
    throw new Error(("Invalid queue priority: " + priority + "."));
  }
  return priority;
}

var tickTask; // tick means the next microtask
var rafTask; // raf means the next animation frame (requestAnimationFrame)
var ricTask; // ric means the next idle perdiod (requestIdleCallback)

var currentTick = Promise.resolve();

function getRaf() {
  return typeof requestAnimationFrame === 'function' ? requestAnimationFrame : setTimeout;
}

function getRic() {
  return typeof requestIdleCallback === 'function' ? requestIdleCallback : getRaf();
}

// schedule a tick task, if it is not yet scheduled
function nextTick(task) {
  if (!tickTask) {
    tickTask = task;
    currentTick.then(runTickTask);
  }
}

function runTickTask() {
  var task = tickTask;
  // set the task to undefined BEFORE calling it
  // this allows it to re-schedule itself for a later time
  tickTask = undefined;
  task();
}

// schedule a raf task, if it is not yet scheduled
function nextAnimationFrame(task) {
  if (!rafTask) {
    rafTask = task;
    var raf = getRaf();
    raf(runRafTask);
  }
}

function runRafTask() {
  var task = rafTask;
  // set the task to undefined BEFORE calling it
  // this allows it to re-schedule itself for a later time
  rafTask = undefined;
  task();
}

// schedule a ric task, if it is not yet scheduled
function nextIdlePeriod(task) {
  if (!ricTask) {
    ricTask = task;
    var ric = getRic();
    ric(runRicTask);
  }
}

function runRicTask() {
  // do not run ric task if there are pending raf tasks
  // let the raf tasks execute first and schedule the ric task for later
  if (!rafTask) {
    var task = ricTask;
    // set the task to undefined BEFORE calling it
    // this allows it to re-schedule itself for a later time
    ricTask = undefined;
    task();
  } else {
    var ric = getRic();
    ric(runRicTask);
  }
}

var TARGET_FPS = 60;
var TARGET_INTERVAL = 1000 / TARGET_FPS;

function queueTaskProcessing(priority) {
  if (priority === priorities.CRITICAL) {
    nextTick(runQueuedCriticalTasks);
  } else if (priority === priorities.HIGH) {
    nextAnimationFrame(runQueuedHighTasks);
  } else {
    nextIdlePeriod(runQueuedLowTasks);
  }
}

function runQueuedCriticalTasks() {
  // critical tasks must all execute before the next frame
  var criticalQueues = queues[priorities.CRITICAL];
  criticalQueues.forEach(processCriticalQueue);
}

function processCriticalQueue(queue) {
  queue.forEach(runTask);
  queue.clear();
}

function runQueuedHighTasks() {
  var startTime = performance.now();
  var isEmpty = processIdleQueues(priorities.HIGH, startTime);
  // there are more tasks to run in the next cycle
  if (!isEmpty) {
    nextAnimationFrame(runQueuedHighTasks);
  }
}

function runQueuedLowTasks() {
  var startTime = performance.now();
  var isEmpty = processIdleQueues(priorities.LOW, startTime);
  // there are more tasks to run in the next cycle
  if (!isEmpty) {
    nextIdlePeriod(runQueuedLowTasks);
  }
}

function processIdleQueues(priority, startTime) {
  var idleQueues = queues[priority];
  var isEmpty = true;

  // if a queue is not empty after processing, it means we have no more time
  // the loop whould stop in this case
  for (var i = 0; isEmpty && i < idleQueues.length; i++) {
    var queue = idleQueues.shift();
    isEmpty = isEmpty && processIdleQueue(queue, startTime);
    idleQueues.push(queue);
  }
  return isEmpty;
}

function processIdleQueue(queue, startTime) {
  var iterator = queue[Symbol.iterator]();
  var task = iterator.next();
  while (performance.now() - startTime < TARGET_INTERVAL) {
    if (task.done) {
      return true;
    }
    runTask(task.value);
    queue.delete(task.value);
    task = iterator.next();
  }
}

function runTask(task) {
  task();
}

var QUEUE = Symbol('task queue');

var Queue = function Queue(priority) {
  this.priority = validatePriority(priority);
  this[QUEUE] = new Set();
  queues[this.priority].push(this[QUEUE]);
};

var prototypeAccessors = { size: {} };

Queue.prototype.has = function has (task) {
  return this[QUEUE].has(task);
};

Queue.prototype.add = function add (task) {
  if (typeof task !== 'function') {
    throw new TypeError((task + " can not be added to the queue. Only functions can be added."));
  }
  var queue = this[QUEUE];
  queue.add(task);
  queueTaskProcessing(this.priority);
};

Queue.prototype.delete = function delete$1 (task) {
  this[QUEUE].delete(task);
};

Queue.prototype.start = function start () {
  var queue = this[QUEUE];
  var priorityQueues = queues[this.priority];
  if (priorityQueues.indexOf(queue) === -1) {
    priorityQueues.push(queue);
  }
  queueTaskProcessing(this.priority);
};

Queue.prototype.stop = function stop () {
  var queue = this[QUEUE];
  var priorityQueues = queues[this.priority];
  var index = priorityQueues.indexOf(queue);
  if (index !== -1) {
    priorityQueues.splice(index, 1);
  }
};

Queue.prototype.clear = function clear () {
  this[QUEUE].clear();
};

prototypeAccessors.size.get = function () {
  return this[QUEUE].size;
};

Queue.prototype.process = function process () {
  var queue = this[QUEUE];
  queue.forEach(runTask);
  queue.clear();
};

Queue.prototype.processing = function processing () {
  var queue = this[QUEUE];
  return new Promise(function (resolve) {
    if (queue.size === 0) {
      resolve();
    } else {
      queue.add(resolve);
    }
  });
};

Object.defineProperties( Queue.prototype, prototypeAccessors );

exports.Queue = Queue;
exports.priorities = priorities;

Object.defineProperty(exports, '__esModule', { value: true });

})));
